# VFS access to the Tcl variable and proc namespaces
#
# 26-09-2000 jcw

package require vfs

package provide nsvfs 0.1

namespace eval nsvfs {
    variable uid 0
    variable types

    namespace export mount umount
}

proc nsvfs::mount {path type args} {
    variable uid 
    variable types
    set ns nsvfs[incr uid]

    switch -- $type {
        vars - procs { set types($ns) $type }
	default {
	    return -code error "type must be either 'vars' or 'procs'"
	}
    }

    ::vfs::mount $path nsvfs::driver $ns

    return $ns
}

proc nsvfs::umount {path args} {
    variable types
    
    foreach {k v} [array get types] {
	if {$v == $path} {unset types($k)}
    }
    
    return [eval [list vfs::umount $path] $args]
}

proc nsvfs::_path {path} {
    variable types
    #tclLog "nsvfs::_path $path"
    set sp [::tcl::file split $path]
    if {$path == "."} {
        set sp {}
    }
    return ::[join $sp ::]
}

proc nsvfs::stat {ns path arr} {
    variable types
    upvar 1 $arr sb
    #tclLog "nsvfs::stat $ns $path $arr"
    set name [_path $path]
    set type $types($ns)

    if {![catch {namespace parent $name}]} {
	set type directory
    } elseif {$type == "vars" && [info exists $name]} {
        set type file
    } elseif {$type == "procs" && [info procs $name] != ""} {
        set type file
    } else {
        return -code error "could not read \"$path\": no such file or directory"
    }
    
    set sb(type)	$type
    set sb(dev)		[list nsvfs::driver $ns]

    if { [string equal $type "directory"] } {
        set sb(atime)   0
        set sb(ctime)   0
	set sb(gid)	0
        set sb(mode)    0777
        set sb(mtime)   0
        set sb(nlink)   [llength [namespace children $name]]
        set sb(size)    0
	set sb(uid)	0
    } else {
        set mtime	[clock seconds]
        set sb(atime)	$mtime
        set sb(ctime)	$mtime
	set sb(gid)	0
        set sb(mode)    0777
        set sb(mtime)	$mtime
        set sb(nlink)   1
	switch $types($ns) {
	    vars { set sb(size) [string length [set $name]] }
	    procs { set sb(size) [string length [_getProc $name]] }
	}
	set sb(uid)	0
    }
}

proc nsvfs::driver {ns option args} {
    variable types
    set type $types($ns)
    #tclLog "nsvfs::driver $ns $option $args"
    switch -- $option {
    lstat       {return [uplevel 1 [concat [list nsvfs::stat $ns] $args]]}
    chdir       {return [lindex $args 0]}
    mtime       -
    delete      -
    stat        -
    getdir      -
    mkdir       {return [uplevel 1 [concat [list nsvfs::$option $ns] $args]]}
    
    open        {
            set file [lindex $args 0]
            set mode [lindex $args 1]
	    set name [_path $file]
	    
            switch -glob -- $mode {
            {}  -
            r   {
		    set fd [memchan]
		    fconfigure $fd -translation binary
		    switch $type {
		        vars { set data [set $name] }
			procs { set data [_getProc $name] }
		    }
		    puts -nonewline $fd $data
		    fconfigure $fd -translation auto
		    seek $fd 0
                    return $fd
                }
            a   {
		    append $name {} ;# make sure it exists
		    set fd [memchan]
                    @close $fd [list nsvfs::do_close $fd $mode $name $type]
                    return $fd
                }
            w   {
		    set $name {}
		    set fd [memchan]
                    @close $fd [list nsvfs::do_close $fd $mode $name $type]
                    return $fd
                }
            default     {
                    error "illegal access mode \"$mode\""
                }
            }
        }
    sync        { }
    umount      { }
    default     {
            return -code error "bad option \"$option\": must be one of chdir, delete, getdir, load, lstat, mkdir, open, stat, sync, or umount"
        }
    }
}

proc nsvfs::do_close {fd mode name type} {
    #tclLog "nsvfs::do_close $fd $mode $name $type"
    flush $fd
    seek $fd 0
    set data [read $fd]
    switch $mode.$type {
	a.vars  { append $name $data }
	w.vars  { set $name $data }
	a.procs { return -code error "can't append to procs" }
	w.procs { _setProc $name $data }
    }
}

proc nsvfs::_getProc {name} {
    set argv {}
    foreach x [info args $name] {
	if {[info default $name $x y]} {
	    set x [list $x $y]
	}
	lappend argv $x
    }
    return [list $argv [info body $name]]
}

proc nsvfs::_setProc {name contents} {
    if {[llength $contents] != 2 || ![info complete $contents]} {
	return -code error "proc $name cannot be saved (bad format?)"
    }
    namespace eval :: [linsert $contents 0 proc $name]
}

proc nsvfs::mkdir {ns path} {
    namespace eval [_path $path] {}
}

proc nsvfs::getdir {ns path {pat *}} {
    variable types
#tclLog "nsvfs::getdir $ns $path $pat"
    set result {}
    foreach x [info $types($ns) [_path $path]::$pat] {
    	lappend result [namespace tail $x]
    }
    foreach x [namespace children [_path $path]] {
	set t [namespace tail $x]
	if [string match $pat $t] {
	    lappend result $t
	}
    }
    return [lsort $result]
}

proc nsvfs::mtime {ns path time} {
    return [clock seconds]
}

proc nsvfs::delete {ns args} {
    variable types
    #tclLog "nsvfs::delete $ns $args"
    set force 0
    while {[llength $args] > 0} {
        set arg [lindex $args 0]
        if { $arg == "-force" } {
            set force 1
            set args [lrange $args 1 end]
        } elseif { $arg == "--" } {
            set args [lrange $args 1 end]
            break
        } else {
            break
        }
    }
    foreach path $args {
        set rc [catch { stat $ns $path sb } err]
        if { $rc }  {
            if { $force } {
                continue
            } else {
                return -code error $err
            }
        }

	set p [_path $path]
        if { $sb(type) == "file" } {
	    switch $types($ns) {
	        vars { unset $p }
		procs { rename $p "" }
	    }
        } else {
            namespace delete $p
        }
    }
    return ""
}

