#
# $Header: /cvsroot/metakit/mk/tcl/kit/kitsh.vfs/lib/vfs/vfs.tcl,v 1.2 2000/12/04 23:50:14 jcw Exp $
#
# Initialization script normally executed in the interpreter for each
# VFS-based application.
#
# Copyright (c) 1999  Matt Newman <matt@sensus.org>
# Further changes made by Jean-Claude Wippler <jcw@equi4.com>
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

# Insist on running with compatible version of Tcl.

package require Tcl 8.2

package provide vfs 0.6
#
# VFS - re-map Tcl filesystem access commands
# to use a virtual table.
#
# vfs.tcl - provides some transparency at the Tcl
# level for standard Tcl I/O and Filesystem routines.
#
# The following commands are re-written in terms of other file commands:
#
#	source
#	file atime
#	file exists
#	file isdirectory
#	file isfile
#	file mtime
#	file size
#	file type
#
# The following commands are overloaded on a per-driver basis:
#
#	file delete
#	file copy
#	file lstat
#	file mkdir
#	file rmdir
#	file stat
#	glob	(file getdir)
#	load	(file load)
#	open	(file open)
#
# The following have been added:
#
#	file getdir <path> ?pat?	- equiv. to cd <path>;glob -nocomplain -- $pat;
#	file open ...		- equiv. to open ...
#	file load ...		- equiv. to load ...
#	file lookup <path>	- returns [list <driver> <stem> <root>]
#	file normalize <path>	- returns absolute path with . and .. resolved
#
# The original versions are available as ::tcl::$cmd
#
namespace eval ::vfs {
    variable debug	0
    if {[info exists env(VFS_DEBUG)]} {
	set debug $env(VFS_DEBUG)
    }

    variable fstab
    variable cwd	[pwd]

    variable temp
    global env

    set temp [file nativename /usr/tmp]
    if {![file exists $temp]} {set temp [file nativename /tmp]}
    catch {set temp $env(TMP)}
    catch {set temp $env(TMPDIR)}
    catch {set temp $env(SYSTEMDRIVE)/temp}
    catch {set temp $env(TEMP)}
    catch {set temp $env(VFS_TEMP)}
    set temp [file join $temp tclkit]
    file mkdir $temp

    # This is not right XXX need somewhere to unpack
    # indirect-dependant DLL's etc.

    global env tcl_platform
    if {$tcl_platform(platform) == "windows"} {
	set env(PATH) "${vfs::temp}/bin;$env(PATH)"
    } elseif {$tcl_platform(platform) == "unix"} {
	set env(PATH) "${vfs::temp}/bin:$env(PATH)"
    } else {
	set env(PATH) "${vfs::temp}/bin"
    }
    proc debug {tag body} {
	set cnt [info cmdcount]
	set time [lindex [time {
	    set rc [catch {uplevel 1 [list eval $body]} ret]
	}] 0]
	set cnt' [info cmdcount]
	set ei ${::errorInfo}
	set ec ${::errorCode}
	puts stderr "$tag: [expr {${cnt'} - $cnt}] ops, $time us"
	return -code $rc -errorcode $ec -errorinfo $ei $ret
    }
}
#
# VFS - re-map Tcl filesystem access commands
# to use a virtual table.
#
# The original versions are available as ::tcl::$cmd
#
#
# Main native Tcl callback
#
namespace eval tcl {

    namespace export file open load source glob close info
}
proc tcl::vfs {option args} {
    switch -- $option {
    sync	{}
    umount	{return -code error "can't unmount native filesystems"}
    chdir	{return [uplevel 1 ::tcl::cd $args]}
    getdir	{return [uplevel 1 ::tcl::getdir $args]}
    open	{return [uplevel 1 ::tcl::open $args]}
    default	{return [uplevel 1 [list ::tcl::file $option] $args]}
    }
}
catch {rename ::cd ::tcl::cd}
catch {rename ::close ::tcl::close}
catch {rename ::file ::tcl::file}
catch {rename ::glob ::tcl::glob}
# Backwards compat - ensure ::tcl::info can be called.
interp alias "" ::tcl::info "" ::info
catch {rename ::load ::tcl::load}
catch {rename ::open ::tcl::open}
catch {rename ::pwd ::tcl::pwd}
catch {rename ::source ::tcl::source}
catch {rename encoding ::tcl::encoding}
#
# Rewrite [source] in terms of open and read.
#
proc ::source {file} {

    ::vfs::log [list source $file]

    set fd [open $file]

    if {[catch {
	fconfigure $fd -eofchar \32
	set data [read $fd]
	close $fd
    } err]} {
	set ec ${::errorCode}
	set ei ${::errorInfo}
	catch {close $fd}
	return -code error -errorinfo $ei -errorcode $ec $err
    }
    # tcl 8.3.1+ only (patched)
    set script [info script]
    info script $file

    set rc [catch {
	uplevel 1 $data
    } err]
    if {$rc == 1} {
	set ei [split ${::errorInfo} \n]
	set eilen [llength $ei]
	incr eilen -3
	set line [lindex $ei $eilen]
	incr eilen -1
	set ::errorInfo [join [lrange $ei 0 $eilen] \n]
	regsub {.*body line ([0-9]*).*} $line {\1} line
	append ::errorInfo "\n    (file \"$file\" line $line)"
    } elseif {$rc == 2} {
	set rc 0
    }
    # tcl 8.3.1+ only (patched)
    info script $script

    return -code $rc -errorinfo ${::errorInfo} -errorcode ${::errorCode} $err
}
#
# Set [close] callbacks
#
proc ::@close {fd args} {
    global @close

    lappend @close($fd) $args
}
#
# Remap [close] so we can fire @close callbacks
#
proc ::close {fd} {
    global @close

    if {[::tcl::info exists @close($fd)]} {
	foreach args [set @close($fd)] {
	    uplevel #0 eval $args
	}
	unset @close($fd)
    }
    ::tcl::close $fd
}
#
# file getdir <dir> ?pat?
#
# Returns files/dirs matching pat in <dir>
#
proc ::tcl::getdir {dir {pat *}} {
    set cwd [pwd]
    cd $dir
    set rc [catch { glob -nocomplain -- $pat } ret]
    cd $cwd

    if {$rc} {
	return ""
    } else {
	return $ret
    }
}
#
# Remap [cd] to go via VFS driver
#
proc ::cd {dir} {
    set rdir [::vfs::normalize $dir]
    if {![file exists $rdir]} {
	return -code error "couldn't change working directory to \"$dir\": no such file or directory"
    } elseif {![file isdirectory $rdir]} {
	return -code error "couldn't change working directory to \"$dir\": not a directory"
    }
    set ::vfs::cwd $rdir
 
    lassign [vfs::lookup $rdir] driver stem root

    eval $driver [list chdir $stem]

    return ""
}
#
# Remap [open] to go via VFS driver
#
proc ::pwd {} {
    return ${::vfs::cwd}
}
#
# Remap [open] to go via VFS driver
#
proc ::open {file {mode r}} {
    ::vfs::log [list open $file $mode]

    if {[string match |* $file]} {
	return [::tcl::open $file $mode]
    } else {
	lassign [vfs::lookup $file] driver stem root

	return [eval $driver [list open $stem $mode]]
    }
}
#
# Remap [file] to go via VFS driver
#
proc ::file {option args} {

    switch -- $option {
	attributes {
	    set file [lindex $args 0]
	    lassign [vfs::lookup [file dirname $file]] driver stem root
	    if {$driver == "::tcl::vfs"} {
		return [uplevel 1 [concat ::tcl::file $option $args]]
	    } else {
		# file is inside vfs
		if {[llength $args] > 1} {
		    set opt [lindex $args 1]
		    switch -- $opt {
			"-longname" {
			   return [file join [file attributes $root -longname] \
			      $stem [file tail $file]]
			}
			"-readonly" {
			    # in principle we can modify any vfs file
			    return 0
			}
			default {
			    # unimplemented
			    error "Unimplemented 'file attributes $file $opt'"
			}
		    }
		} else {
		    # just 'file attributes $file'
		    foreach opt [list -longname -readonly] {
			lappend res $opt [file attributes $file $opt]
		    }
		    return $res
		}
	    }
	}
	readable	{return [file exists [lindex $args 0]]}
	lookup	{return [eval vfs::lookup $args]}
	getdir	{ 
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root
	    return [eval $driver [list $option $stem]]
	}
	atime	-
	isdirectory	{
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root

	    if {[catch {
		eval $driver [list stat $stem sb]
	    }] || $sb(type) != "directory"} {
		return 0
	    } else {
		return 1
	    }
	}
	isfile	{
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root

	    if {[catch {
		eval $driver [list stat $stem sb]
	    }] || $sb(type) != "file"} {
		return 0
	    } else {
		return 1
	    }
	}
	mtime	{
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root

	    if {[llength $args] > 1} {
		eval $driver [list mtime $stem [lindex $args 1]]
		return
	    } else {
		eval $driver [list stat $stem sb]
		return $sb($option)
	    }
	}
	size	-
	type	{
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root

	    eval $driver [list stat $stem sb]

	    return $sb($option)
	}
	exists	{
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root

	    if {[catch [concat $driver [list stat $stem sb]]]} {
		return 0
	    } else {
		return 1
	    }
	}
	lstat	-
	stat	{
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root

	    upvar 1 [lindex $args 1] sb

	    set sb(vfs) $driver
	    set sb(vroot) $root
	    set sb(vstem) $stem

	    return [eval $driver [list stat $stem sb]]
	}
	mkdir	-
	rmdir	-
	delete	{
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root

	    return [eval $driver [list $option $stem]]
	}
	rename -
	copy	{
	    set start 0
	    foreach opt $args {
		switch -- $opt {
		    "--" {
			break
		    }
		    "-force" {
			set force 1
		    }
		    default {
			break
		    }
		}
		incr start
	    }
	    set args [lrange $args $start end]
	    if {[llength $args] < 2} {
		error "wrong # args: should be \"file $option ?options? source ?source
    ..? target\""
	    }
	    set dst [lindex $args end]
	    lassign [vfs::lookup $dst] dst-driver dst-stem dst-root
	    # How do we handle '-force' ?
	    foreach src [lrange $args 0 end-1] {
		lassign [vfs::lookup $src] src-driver src-stem src-root
		# neither 'copy' nor 'rename' are implemented in mk4vfs
		if {${src-driver} == ${dst-driver} &&  ${src-root} == ${dst-root} &&
    ([lindex ${src-driver} 0] != "mk4vfs::driver")} {
		    # Inside same "filesystem"
		    eval ${src-driver} [list $option $src $dst]
		} else {
		    # Do a cross-filesystem copy
		    # beware - if src is a directory then a deep-copy is performed.
		    # Four options exist:
		    # src=file	dst=file (exist or not)
		    # src=file	dst=dir
		    # src=dir	dst=!exists (created as dir)
		    # src=dir	dst=dir
		    #
		    set sd [file isdirectory $src]
		    if {![file exists $dst]} {
			if {$sd} {
			    file mkdir $dst
			}
		    } else {
			if {[file isdirectory $dst]} {
			    set dst [file join $dst [file tail $src]]
			}
		    }
		    if {!$sd} {
			set sfd [open $src]
			fconfigure $sfd -translation binary
			set dfd [open $dst w]
			fconfigure $dfd -translation binary
			
			fcopy $sfd $dfd
			close $sfd
			close $dfd
			file mtime $dst [file mtime $src]
		    } else {
			error "deep $option not implemented (yet)"
		    }
		    if {$option == "rename"} {
			eval ${src-driver} [list delete ${src-stem}]
		    }
		}
	    }
	}
	Xcopy	{ # original code left in for now, "Xcopy" is never used
	    set src [lindex $args 0]
	    set dst [lindex $args 1]

	    lassign [vfs::lookup $src] src-driver src-stem src-root
	    lassign [vfs::lookup $dst] dst-driver dst-stem dst-root

	    if {${src-driver} == ${dst-driver} && \
		${src-root} == ${dst-root} } {
		# Inside same "filesystem"
		return [eval ${src-driver} [list copy $src $dst]]
	    }
	    # Do a cross-filesystem copy
	    # beware - if src is a directory then a deep-copy is performed.
	    # Four options exist:
	    # src=file	dst=file (exist or not)
	    # src=file	dst=dir
	    # src=dir	dst=!exists (created as dir)
	    # src=dir	dst=dir
	    #
	    set sd [file isdirectory $src]
	    if {![file exists $dst]} {
		if {$sd} {
		    file mkdir $dst
		}
	    } else {
		if {[file isdirectory $dst]} {
		    set dst [file join $dst [file tail $src]]
		}
	    }
	    if {!$sd} {
		set sfd [open $src]
		fconfigure $sfd -translation binary
		set dfd [open $dst w]
		fconfigure $dfd -translation binary

		fcopy $sfd $dfd
		close $sfd
		close $dfd
		file mtime $dst [file mtime $src]
	    } else {
		error "deep copy not implemented (yet)"
	    }
	}
	normalize {
	    set file [lindex $args 0]
	    return [::vfs::normalize $file]
	}
	sync	{
	    set file [lindex $args 0]
	    lassign [vfs::lookup $file] driver stem root

	    eval $driver [list $option $stem] [lrange $args 1 end]
	}
	default	{
	    #::vfs::log "tcl::file $option $args"
	    return [uplevel 1 [concat ::tcl::file $option $args]]
	}
    }
}
#
# Remap [load] to go via VFS driver
#
proc ::load {path {pkg ""} {interp ""}} {
    #::vfs::log "load $path $pkg $interp"
    # If the stat fails we fall-through to standard load, as
    # library may be on our path.

    if {[catch {file stat $path sb}] || \
	$sb(vfs) == "::tcl::vfs"} {
	::vfs::log [list tcl::load $path $pkg $interp]
	return [uplevel 1 [list ::tcl::load $path $pkg $interp]]
    }
    # See if it is already loaded?
    if {![catch [list uplevel 1 [list ::tcl::load {} $pkg $interp]]]} {
	::vfs::log [list tcl::load {} $pkg $interp]
	return
    }
    set fileName [vfs::unpack $path]
    catch {file attributes $fileName -permissions 0777}
    ::vfs::log [list tcl::load $fileName $pkg $interp]
    uplevel 1 [list ::tcl::load $fileName $pkg $interp]
}
#
# Remap [glob] to go via VFS driver
#
proc ::glob {args} {

    ::vfs::log "glob $args"

    set complain 1
    set join 0
    set onlydirs 0
    while {[llength $args] > 0} {
	set opt [lindex $args 0]
	set args [lrange $args 1 end]
	switch -glob -- $opt {
	    -- {
		break
	    }
	    -n* { # -nocomplain
	    	set complain 0
	    }	
	    -j* { # -join
	    	set join 1
	    }	
	    -d* { # -dir <dir>
		set dir [lindex $args 0]
		set args [lrange $args 1 end]
	    }
	    -p* { # -path <path>
		set path [lindex $args 0]
		set args [lrange $args 1 end]
	    }
	    -t* { # -types <types>
		set types [lindex $args 0]
		if {[lsearch -exact $types d] >= 0} {
		    set onlydirs 1
		}    
		set args [lrange $args 1 end]
	    }
	    default {
	    	set args [linsert $args 0 $opt]
		break
	    }
	}    
    }	
    
    regexp {Z(.)Z} [file join Z Z] "" separator
    if {$join} {
	set args [list [eval file join $args]]
    }                                                                   
    
    set add ""
    if {$onlydirs} {
	set add $separator
    }

    if {[info exists path]} {
	if {[info exists dir]} {
	    if {$complain} {
		error "Can't use option '-dir' with '-path'"
	    }
	    return ""
	}
	regsub -all {[][*?\{\}\\]} $path {\\&} prefix
    } elseif {[info exists dir]} {
	regsub -all {[][*?\{\}\\]} $dir {\\&} prefix
	append prefix $separator
    } else {
	set prefix ""
    }                                                                   

    foreach pat $args {
	foreach exp [::vfs::ExpandGlob $prefix$pat$add] {
#::vfs::log exp=$exp 1
	    set rc [catch {
		_glob "" [::tcl::file split $exp] hits $complain
	    } err]
	    if {$rc && $complain } {
		return -code error -errorinfo ${::errorInfo} $err
	    }
	}
    }

    if {$onlydirs} {
	foreach hit [array names hits] {
	    file stat $hit sb
	    if {$sb(type) != "directory"} {
		unset hits($hit)
	    }
	}
    }

    # serialize result
    set result [lsort [array names hits]]

    if {[array size hits] == 0 && $complain} {
	set s ""
	if {[llength $args] > 1} { set s s }
	return -code error "no files matched glob pattern$s \"$args\""
    }
    ::vfs::log "glob => $result" 1
    return $result
}
proc ::_glob {pstem plist arr complain} {
    upvar 1 $arr hits

    #::vfs::log [list _mg = $pstem $plist $arr $complain] 1

    set pat [lindex $plist 0]
    set plist [lrange $plist 1 end]
    set path [file join $pstem $pat]

    if {![regexp {[\*?]} $pat]} {
	if {[llength $plist] > 0} {
	    _glob $path $plist hits $complain
	    return
	}
    }
    lassign [vfs::lookup $path] driver stem root

    set dp [file tail $stem]
    if {$dp == ""} {set dp *}
    ::vfs::log "trying [concat $driver [list getdir [file dirname $stem] $dp]]" 1
    set rc [catch {
	set matches [eval $driver [list getdir [file dirname $stem] $dp ]]
    } err]
    if {$rc} {
	if {$complain} {
	    return -code error -errorinfo ${::errorInfo} $err
	}
	set matches {}
    }

    if {[llength $plist] == 0} {
	foreach hit $matches {
	    set match [file join $pstem [file tail $hit]]
	    set hits($match) 1
	}
    } else {
	foreach hit $matches {
	    #tclLog "$stem => $hit"
	    set match [file join $pstem [file tail $hit]]
	    _glob $match $plist hits $complain
	}
    }
}

# code to circumvent the glob in the encodings C code when VFS-based
# this code gratefully accepted from Vince Darley <vince.darley@eurobios.com>
proc encoding {cmd args} {
     if {$cmd == "names"} {
	set names [::tcl::encoding names]
	foreach enc [glob -nocomplain -dir [file join [info library] encoding] *.enc] {
	    set name [file rootname [file tail $enc]]
	    if {[lsearch -exact $names $name] == -1} {
		lappend names $name
	    }
	}
	return $names
     } else {
	uplevel 1 [list ::tcl::encoding $cmd] $args
     }
}

proc vfs::log {msg {lvl 0}} {
    if {$lvl < ${::vfs::debug}} {
	puts stderr "vfs($lvl): $msg"
    }
}

if 0 { # original version
proc vfs::lookup {file} {
    variable fstab

#vfs::debug vfs::lookup($file) {}
    set sp [::tcl::file split [::vfs::normalize $file]]
    # match this against the most specific mount-point
    set driver ::tcl::vfs
    set root $sp
    set stem ""
    set ridx [expr {[llength $root]-1}] 
    
    while {[llength $root]} {
	set path [eval ::tcl::file join $root]
#log "test $path"
        if {[::tcl::info exists fstab($path)]} {
	    set driver $fstab($path)
	    break
	}
	set stem [linsert $stem 0 [lindex $root end]]
	set root [lrange $root 0 [incr ridx -1]]
    }
    if {[llength $root] > 0} {
	set nroot [eval ::tcl::file join $root]
    } else {
	set nroot ""
    } 
    if {[llength $stem] > 0} {
	set nfile [eval ::tcl::file join $stem]
    } else {
	set nfile {}
    }
#log "vfs mount point: ${driver}:$root -> $stem"
    return [list $driver $nfile $nroot]
}
} ;# replaced by Vince Darley's version:

proc vfs::lookup {file} {
    variable fstab

#vfs::debug vfs::lookup($file) {}
    set root [::tcl::file split [::vfs::normalize $file]]
    # match this against the most specific mount-point
    set ridx [llength $root]
    set cmd [list ::tcl::file join]
    
    while {$ridx > 0} {
	incr ridx -1
	set nroot [eval $cmd [lrange $root 0 $ridx]]
#log "test $nroot"
	if {[::info exists fstab($nroot)]} {
	    set driver $fstab($nroot)
	    break
	}
    }
    if {![::info exists driver]} {
	set driver ::tcl::vfs
	set nroot ""
	set nfile $root
    } else {
	set nfile [lrange $root [expr {$ridx +1}] end]
    }
    if {[llength $nfile] > 0} {
	set nfile [eval $cmd $nfile]
    }
#log "vfs mount point: ${driver}:$nroot -> $nfile"
    return [list $driver $nfile $nroot]
}

proc ::vfs::mount {args} {
    variable fstab

    ::vfs::log "mount $args"

    if {[llength $args] == 0} {
	return [array get fstab]
    } else {
	set path [::vfs::normalize [lindex $args 0]]
	set fstab($path) [lrange $args 1 end]
    }
}

proc ::vfs::umount {path args} {
    set path [::vfs::normalize $path]
    variable fstab

    ::vfs::log "umount $path $args"

    if {![info exists fstab($path)]} {
	return -code error "mount point \"$path\" does not exist"
    }
    eval $fstab($path) umount $args
    unset fstab($path)
}

proc ::vfs::auto {dir args} {
    variable fstab

    set np {}
    foreach ele [::tcl::file split $dir] {
	lappend np $ele
	switch -- [string tolower [file extension $ele]] {
	.tkd	-
	.cf4	{
		package require mk4vfs

		set path [::vfs::normalize [eval file join $np]]
		if {![info exists fstab($path)]} {
		    ::vfs::log "automounting $path"
		    eval [list mk4vfs::mount $path $path] $args
		}
		return
	    }
	.zip	{
		package require vfs.zip

		set path [eval file join $np]
		set path [::vfs::normalize [eval file join $np]]
		if {![info exists fstab($path)]} {
		    ::vfs::log "automounting $path"
		    zip::mount $path $path
		}
		return
	    }
	.exe	{
		set path [::vfs::normalize [eval file join $np]]
		if {[info exists fstab($path)]} {
		    continue
		}
		if { [catch {mk::file end $path}] } {
		    ::vfs::log "automounting $path (driver=zip)"
		    zip::mount $path $path
		} else {
		    ::vfs::log "automounting $path (driver=mk)"
		    eval [list mk4vfs::mount $path $path] $args
	        }
	    }
	}
    }
}

proc ::vfs::ExpandGlob {pat} {
#::vfs::log "expand $pat"
    #
    # Convert FILE GLOB-style pattern to 
    # standard string match-style pattern
    #
    # i.e. expand ...{a,b,c} constructs
    #
    set f [string first "\{" $pat]
    if {$f == -1} {
	return [list $pat]
    }
    set pos [string range $pat [expr $f+1] end]
    set head [string range $pat 0 [expr {$f-1}]]

    set l [string first "\}" $pos]
    set tail [string range $pos [expr {$l+1}] end]

    set pos [string range $pos 0 [expr {$l-1}]]

    set result {}
    foreach bit [split $pos ,] {
	lappend result "$head$bit$tail"
    }
    return $result
}
#
# This is quite tricky - we need to unpack files
# from a VFS into a native FS location. This needs
# to be repeatable, lazy and *robust*.
#
proc ::vfs::unpack {path} {
    package require Trf	;# For md5

    file stat $path sb

    if {$sb(vroot) == ""} {# Native - don't unpack
	return $path
    }
    # Generate a sig for virtual root
    binary scan [md5 [file dirname $sb(vroot)]] H* sig
    set sig [file join [string range $sig 0 15] [string range $sig 16 end]]

    set fileName [file join ${::vfs::temp} $sig $sb(vstem)]
    set rc [catch {file stat $fileName pb}]
    if {$rc || $sb(size) != $pb(size) || $sb(mtime) > $pb(mtime)} {

	::vfs::log "unpacking $path -> $fileName"

	mkdir [file dirname $fileName]

	set f [open $path]
	set g [open $fileName w]
	fconfigure $f -translation binary -encoding binary
	fconfigure $g -translation binary -encoding binary
	fcopy $f $g
	close $f
	close $g
    } else {
	::vfs::log "not unpacking $path -> $fileName"
    }
    return $fileName
}
proc ::vfs::mkdir {path} {
    # SLOW but IMPORTANT - we *must* make each directory writable
    # in a multi-user system such as unix this *IS* a big deal
    set sp {}
    foreach ele [::tcl::file split $path] {
	lappend sp $ele
	set path [eval ::tcl::file join $sp]
	if {![::tcl::file isdirectory $path]} {
	    ::tcl::file mkdir $path
	    catch {file attributes $path -permissions 0777}
	}
    }
}
if { [llength [info commands lassign]] == 0 } {
    proc lassign {list args} {
	foreach val $list var $args {
	    uplevel 1 [list set $var $val]
	}
    }
}
