// Copyright (c) 1999-2000 Jean-Claude Wippler <jcw@equi4.com>
//
//  Tcl parser

#include "wcommon.h"

#include <ctype.h>
#include <stdlib.h>     // strtol

/////////////////////////////////////////////////////////////////////////////

class TclEngine
{
    wSeqRef _stack;
    const char* _start;

    wSeqRef LitResult(char type_, const char* from_, const char* to_);

public:
    TclEngine (const char* _start =0);
    wRef TclScan(const char* from_, const char* to_);
    wRef TclExpr(const char* from_, const char* to_);

    void LitChar(char ch_);
    void Literal(const char* from_, const char* to_);
    int Frame();
    void Command(int base_);
    int Merge(int wbase_);
    wRef Pop();
    void Scalar(const char* from_, const char* to_);
    void Array(const char* from_, const char* to_);
    void Quote(bool unbalanced_);
};

/////////////////////////////////////////////////////////////////////////////

class TclScanner
{
protected:
    const char* _from;
    const char* _to;

    // Look ahead to the next character
    char PeekChar() const
    {
        return _from != _to ? *_from : 0;
    }

    // Character classes
    enum eClass {
        eNormal       = 0x00,
        eSpace        = 0x01,
        eNewLine      = 0x02,
        eSemiColon    = 0x04,
        eSubstitute   = 0x08,
        eDoubleQuote  = 0x10,
        eCloseParen   = 0x20,
        eCloseBracket = 0x40,
        eBrace        = 0x80
    };

    // Determine the character class
    static eClass CharClass(char c)
    {
        return  c == ' '  || c == '\t' ?            eSpace :
                c == '\n' ?                         eNewLine :
                c == ';' || c == 0 ?                eSemiColon :
                c == '\\' || c == '$' || c == '[' ? eSubstitute :
                c == '"' ?                          eDoubleQuote :
                c == ')' ?                          eCloseParen :
                c == ']' ?                          eCloseBracket :
                c == '{' || c == '}' ?              eBrace :
                                                    eNormal;
    }

    // Return next character, or a null byte if at end
    eClass PeekClass() const
    {
        return CharClass(PeekChar());
    }

    // Return non-zero if next character is white space or newline
    int AtWhiteOrEol() const
    {
        return PeekClass() & (eSpace | eNewLine);
    }

    // Returns number of chars which can be replace by a space, or zero
    int EscapedNewLine() const
    {
        if (_from + 1 >= _to || _from[0] != '\\' || _from[1] != '\n')
            return 0;

        int i = 2;
        while (_from + i < _to && CharClass(_from[i]) == eSpace)
            ++i;
        return i;
    }

    // Skip over whitespace and escaped newlines
    void SkipWhiteSpace()
    {
        for (;;)
        {
            while (PeekClass() & eSpace)
                ++_from;

            int i = EscapedNewLine();
            if (i == 0)
                return;

            _from += i;
        }
    }

public:
    // Define a new scanner over a specified string
    TclScanner (const char* from_, const char* to_) 
        : _from (from_), _to (to_)
    {
    }

    char ConvertBackslash();

        enum eFindResult {
                efrEof = -1,
                efrLiteral,
                efrEscaped,
                efrExtraAfterBrace,
                efrMissingBrace,
                efrExtraAfterQuote,
                efrMissingQuote
        };
        
    eFindResult FindListElement(const char*& begin_, const char*& end_);
    
    int ConvertListElement(const char* begin_, const char* end_, char* buffer_);
};

char TclScanner::ConvertBackslash()
{
    if (_from == _to)
        return '\\';

    const char* start = _from++;
    
    if (*start == 'x' && isxdigit(PeekChar()))
    {
        char ch = 0;

        while (isxdigit(PeekChar()))
        {
            ch = (char) (16 * ch + (*_from & 0x1F));
            if (*_from++ > '9')
                ch += 9;
        }

        return ch;
    }
    
    if (isdigit(*start))
    {
        char ch = 0;

        while (isdigit(PeekChar()))
            ch = (char) (8 * ch + *_from++ - '0');
        
        return ch;
    }
    
    if (*start == '\n')
    {
        _from = start - 1;
        _from += EscapedNewLine();
        return ' ';
    }

    char ch = *start;

    switch (ch)
    {
        case 'a':   ch = 0x07; break;
        case 'b':   ch = 0x08; break;
        case 'f':   ch = 0x0C; break;
        case 'n':   ch = 0x0A; break;
        case 'r':   ch = 0x0D; break;
        case 't':   ch = 0x09; break;
        case 'v':   ch = 0x0B; break;
    }

    return ch;
}

// Find the boundaries of the next list element
TclScanner::eFindResult TclScanner::FindListElement(const char*& begin_, const char*& end_)
{
    int level = 0;
    eFindResult escaped = efrLiteral;

    while (AtWhiteOrEol())
        ++_from;

    if (_from == _to)
        return efrEof;

    switch (PeekChar())
    {
        case '{':
            begin_ = ++_from;
        
            while (_from != _to)
            {
                switch (*_from++)
                {
                case '{':
                    ++level;
                    break;

                case '}':
                    if (--level >= 0)
                        break;

                    end_ = _from - 1;
                    return AtWhiteOrEol() || _from == _to ? escaped :
                                efrExtraAfterBrace; // extra characters after close-brace

                case '\\':
                    if (_from != _to)
                        ++_from;
                    escaped = efrEscaped;
                }
            }
            return efrMissingBrace; // missing closing brace in list element

        case '"':
            begin_ = ++_from;

            while (_from != _to)
            {
                switch (*_from++)
                {
                case '"':
                    end_ = _from - 1;
                    return AtWhiteOrEol() || _from == _to ? escaped :
                                efrExtraAfterQuote; // extra characters after close-quote

                case '\\':
                    if (_from != _to)
                        ++_from;
                    escaped = efrEscaped;
                }
            }
            return efrMissingQuote; // missing quote in list element

        default:
            begin_ = _from;

            do
            {
                if (*_from == '\\')
                {
                    if (++_from == _to)
                        break;

                    escaped = efrEscaped;
                }
            } while (!AtWhiteOrEol() && ++_from != _to);

            end_ = _from;
            return escaped;
    }
}

int TclScanner::ConvertListElement(const char* begin_, const char* end_, char* buffer_)
{
    char* p = buffer_;

    const char* save = _from;
    _from = begin_;

    while (_from != end_)
    {
        char c = *_from++;
        if (c == '\\' && _from != end_)
            c = ConvertBackslash();
        *p++ = c;
    }

    _from = save;
    return p - buffer_;
}

class TclTokenizer : public TclScanner
{
protected:
    TclEngine& _engine;

    // Fail and throw an error if next char is incorrect, else skip it
    void Require(char ch_, const char* error_)
    {
        if (PeekChar() != ch_)
            throw error_;
        ++_from;
    }

public:
    // Define a new tokenizer over a specified string
    TclTokenizer (const char* from_, const char* to_, TclEngine& engine_) 
        : TclScanner (from_, to_), _engine (engine_)
    {
    }

    // The number of unparsed characters remaining
    size_t Remaining() const
    {
        return _to - _from;
    }

    void TokenizeScript();
    void TokenizeCommand(bool nested_);
    bool TokenizeUntil(int mask_);
    bool TokenizeVariable();
};

// Tokenize all the commands
void TclTokenizer::TokenizeScript()
{
    while (_from != _to)
    {
        TokenizeCommand(false);
    }
}
// Tokenize the next command, end at close bracket is nested
void TclTokenizer::TokenizeCommand(bool nested_)
{
    const char* init = _from;

    for (;;)
    {
        SkipWhiteSpace();

        if (PeekClass() & (eNewLine | eSemiColon))
        {
            if (_from == _to)
                return;

            ++_from;
            continue;
        }

        if (PeekChar() != '#')
            break;

        while (++_from != _to)
        {
            // this is non-conformant: no escaped backslashes
            if (*_from == '\n' && _from[-1] != '\\')
                break;
        }
    }

    if (_from == _to)
        return;

    int base = _engine.Frame();
    int wbase = base;

    int mask = eNewLine | eSemiColon;
    if (nested_)
        mask |= eCloseBracket;

    const char* first = _from;

    while (_from != _to)
    {
        const char* start = _from++;

        if (*start == '"')
        {
            TokenizeUntil(eDoubleQuote);
            Require('"', "missing close-quote");
//          if (_from != _to && !AtWhiteOrEol() && nested_ && *_from != ']')
//              throw TclException ("extra characters after close-quote");
        }
        else if (*start == '{')
        {
            const char* token = _from;

            int level = 0;
            for (;;)
            {
                while (PeekClass() == eNormal)
                    ++_from;
                
                if (_from == _to)
                    break;

                if (*_from == '\\' && _from + 1 != _to)
                {
                    if (_from[1] == '\n')
                    {
                        if (_from > token)
                            _engine.Literal(token, _from);

                        _from += EscapedNewLine();
                        _engine.LitChar(' ');

                        token = _from;
                    }
                    else
                        ++_from;
                }
                else if (*_from == '{')
                    ++level;
                else if (*_from == '}' && --level < 0)
                    break;

                ++_from;
            }

            if (_from > token)
                _engine.Literal(token, _from);

            Require('}', "missing close-brace");
//          if (_from != _to && !AtWhiteOrEol() && nested_ && *_from != ']')
//              throw TclException ("extra characters after close-brace");
        }
        else
        {
            --_from;
            TokenizeUntil(mask | eSpace);
        }

        const char* end = _from;
        SkipWhiteSpace();

        if (PeekClass() & mask)
            break;

        wbase = _engine.Merge(wbase);
    }

    _engine.Merge(wbase);
    _engine.Command(base);
}

// Tokenize a number of tokens until specific characters are reached
bool TclTokenizer::TokenizeUntil(int mask_)
{
    while (_from != _to)
    {
        const char* start = _from;

        if (CharClass(*_from) & mask_)
            return true;

        switch (*_from++)
        {
            case '$':
                if (!TokenizeVariable())
                    _engine.Literal(start, _from);
                break;

            case '[':
                while (_from != _to && *_from != ']')
                    TokenizeCommand(true);

                Require(']', "missing close-bracket");
                break;

            case '\\':
                --_from;

                Require('\\', "missing \\");
                _engine.LitChar(ConvertBackslash());

                if (start[1] == '\n' && (mask_ & eSpace))
                    return false;

                break;

            default:
                while (_from != _to &&
                        (CharClass(*_from) & (mask_ | eSubstitute)) == 0)
                    ++_from;

                _engine.Literal(start, _from);
        }
    }

    return true;
}

// Tokenize veriable names, in all their variations
bool TclTokenizer::TokenizeVariable()
{
    if (PeekChar() == '{')
    {
        const char* start = ++_from;

        while (_from != _to && *_from++ != '}')
            ;

        --_from;
        Require('}', "missing close-brace for variable name");

        _engine.Scalar(start, _from - 1);
    }
    else
    {
        const char* start = _from;

        while (_from != _to)
        {
            if (!isalnum(*_from) && *_from != '_')
                break;
            ++_from;
        }

        if (PeekChar() != '(')
        {
            if (start == _from)
                return false; // not a variable after all

            _engine.Scalar(start, _from);
        }
        else
        {
            const char* end = _from;

            ++_from;
            TokenizeUntil(eCloseParen);

            Require(')', "missing )");

            _engine.Array(start, end);
        }
    }

    return true;
}

/////////////////////////////////////////////////////////////////////////////

	// moved out of inline code, because some compilers don't like it
    static const char* precedences[] =
	{ "o", "a", "|", "^", "&", "en", "<>lg", "ud", "+-", "*/%", 0 };

class TclExprScanner : public TclTokenizer
{
    char _token;
    bool _valid;
    wRef _value;

    void LookAhead(const char* chars_)
    {
        while (*chars_)
            if (PeekChar() == *chars_++)
            {
                _token = *chars_;
                ++_from;
                break;
            }
            else
                ++chars_;
    }

    char Peek()
    {
        if (!_valid)
        {
            while (AtWhiteOrEol())
                ++_from;

            _valid = true;
            _value = wRef ();

            if (_from == _to)
            {
                _token = 0;
            }
            else
            {
                _token = *_from++;

                switch (_token)
                {
                    case '&':   LookAhead("&a");    break;
                    case '|':   LookAhead("|o");    break;

                    case '<':   LookAhead("<u=l");  break;
                    case '>':   LookAhead(">d=g");  break;
                    case '=':   LookAhead("=e");    break;
                    case '!':   LookAhead("=n");    break;

                    case '$':   if (TokenizeVariable())
                                {
                                    _token = 'V';
                                    _value = _engine.Pop();
                                }
                                break;

                    case '[':   while (_from != _to && *_from != ']')
                                    TokenizeCommand(true);

                                Require(']', "missing close-bracket");

                                _token = 'V';
                                _value = _engine.Pop();
                                break;

                    case '"':   TokenizeUntil(eDoubleQuote);
                                Require('"', "missing \"");

                                _token = 'V';
                                _value = _engine.Pop();
                                break;

                    default:    if (isdigit(_token))
                                {
                                    const char* start = --_from;
                                        //! hm, strtol needs a terminating character
                                    strtol(_from, (char**) &_from, 0);
                                    _engine.Literal(start, _from);
                                    _token = 'V';
                                    _value = _engine.Pop();
                                }
                }
            }
        }

        return _token;
    }

    char Consume()
    {
        char token = Peek();
        _valid = false;
        return token;
    }

    wRef TopLevel()
    {
        wRef result = Descend(0);
        if (Peek() == '?')
        {
            Consume();
            wRef trueVal = TopLevel();
            Require(':', "missing colon in conditional expression");
            if (Consume() != ':')
                throw /*TclException*/ ("missing colon in conditional expression");
            wRef falseVal = TopLevel();
            result = (int) result ? trueVal : falseVal;
        }
        return result;
    }

    wRef Descend(int level_)
    {
        const char* ops = precedences[level_];
        if (ops == 0)
            return Term();

        wRef result = Descend(level_ + 1);
        //printf("Descend level_ %d value %ld\n", level_, result.GetInt());
        while (*ops)
        {
            if (Peek() == *ops)
            {
                char op = Consume();
                BinOp(result, op, Descend(level_ + 1));
                ops = precedences[level_]; // restart
            }
            else
                ++ops;
        }

        return result;
    }

    wRef Term()
    {
        wSeqRef result;
        char token = Consume();

        switch (token)
        {
            case '(':   result = TopLevel();
                        if (Consume() != ')')
                            throw /*TclException*/ ("left parenthesis missing");
                        break;

            case 'V':   result = _value;
                        break;

            case '+': case '-': case '~': case '!':
                    {
                        result.Append("U");
                        result.Append(wRef (&token, 1));
                        result.Append(Term());
                        break;
                    }

            default:    throw /*TclException*/ ("malformed expression");
        }

        return result;
    }

    void BinOp(wRef& left_, char op_, const wRef& right_)
    {
        wSeqRef r;
        r.Append("B");
        r.Append(wRef (&op_, 1));
        r.Append(left_);
        r.Append(right_);
        left_ = r;
    }

public:
    TclExprScanner (const char* from_, const char* to_, TclEngine& engine_) 
        : TclTokenizer (from_, to_, engine_), _valid (false)
    {
    }

    wRef Evaluate()
    {
        wRef result = TopLevel();
        if (Peek() != 0)
            throw /*TclException*/ ("spurious characters after expression");
        return result;
    }
};

wRef ListItem(const char* from_, const char* to_)
{
    int braceCount = 0;

    bool useBraces = false;
    bool avoidBraces = false;
    bool unBalanced = false;

    if (from_ == to_ || *from_ == '{' || *from_ == '"')
        useBraces = true;

    for (const char* p = from_; p != to_; ++p)
    {
        switch (*p)
        {
    case '{':   ++braceCount;
                break;

    case '}':   if (--braceCount < 0)
                    avoidBraces = unBalanced = true;
                break;

    case '[': case '$': case ';': case ' ':
    case '\f': case '\n': case '\r': case '\t': case '\v':
                useBraces = true;
                break;

    case '\\':  if (p + 1 == to_ || p[1] == '\n')
                    avoidBraces = unBalanced = true;
                else
                    useBraces = true;
                break;
        }
    }

    if (braceCount != 0)
        avoidBraces = unBalanced = true;

    TclEngine engine;
    engine.Literal(from_, to_);

    if (!avoidBraces) {
        if (useBraces)
        {
            int base = engine.Frame();
            engine.LitChar('{');
            if (from_ != to_)
                engine.Literal(from_, to_);
            engine.LitChar('}');
            engine.Merge(base);
            return engine.Pop();
        }
        if (!unBalanced)
            return engine.Pop();
    }

    engine.Quote(unBalanced);
    return engine.Pop();
}

wRef QuoteItem(const char* from_, const char* to_, bool unbalanced_)
{
    char* buffer = new char [2 * (to_-from_)];

    char* q = buffer;
    while (from_ != to_)
    {
        char c = *from_++;

        switch (c) {
    case '[': case ']': case '$': case ';':
    case ' ': case '\\': case '"':
                *q++ = '\\';
                break;
    
    case '{': case '}':
                if (unbalanced_)
                    *q++ = '\\';
                break;

    case '\f':  *q++ = '\\'; c = 'f'; break;
    case '\n':  *q++ = '\\'; c = 'n'; break;
    case '\r':  *q++ = '\\'; c = 'r'; break;
    case '\t':  *q++ = '\\'; c = 't'; break;
    case '\v':  *q++ = '\\'; c = 'v'; break;
        }

        *q++ = c;
    }

    wRef result (buffer, q - buffer);
    delete [] buffer;
    return result;
}

/////////////////////////////////////////////////////////////////////////////

TclEngine::TclEngine (const char* start_)
    : _start (start_)
{
}

wSeqRef TclEngine::LitResult(char type_, const char* from_, const char* to_)
{
    wSeqRef r;
    r.Append(wRef (&type_, 1));
    if (_start != 0)
    {
        r.Append(from_ - _start);
        r.Append(to_ - from_);
    }
    else
        r.Append(wRef (from_, to_ - from_));
    return r;
}

void TclEngine::LitChar(char ch_)
{
    wSeqRef r;
    r.Append("C");
    r.Append((int) ch_);
    _stack.Append(r);
}

void TclEngine::Literal(const char* from_, const char* to_)
{
    if (from_ + 1 == to_ && *from_ <= ' ' || *from_ > '~')
        LitChar(*from_);
    else
        _stack.Append(LitResult('L', from_, to_));
}

int TclEngine::Frame()
{
    return _stack.Size();
}

void TclEngine::Command(int base_)
{
    int n = _stack.Size() - base_;
    wSeqRef r = _stack.GetSlice(base_, _stack.Size());
    _stack.Delete(base_, n);
    r.Insert(0, "X");
    _stack.Append(r);
}

int TclEngine::Merge(int wbase_)
{
    int n = _stack.Size() - wbase_;
    if (n != 1)
    {
        wSeqRef r = _stack.GetSlice(wbase_, _stack.Size());
        _stack.Delete(wbase_, n);
        r.Insert(0, "M");
        _stack.Append(r);
    }
    return _stack.Size();
}

wRef TclEngine::Pop()
{
    wRef r;
    int n = _stack.Size() - 1;
    if (n >= 0)
    {
        r = _stack[n];
        _stack.Delete(n);
    }
    return r;
}

void TclEngine::Quote(bool unbalanced_)
{
    wSeqRef r;
    r.Append(unbalanced_ ? "U" : "B");
    r.Append(Pop());
    _stack.Append(r);
}

void TclEngine::Scalar(const char* from_, const char* to_)
{
    _stack.Append(LitResult('S', from_, to_));
}

void TclEngine::Array(const char* from_, const char* to_)
{
    wSeqRef r = LitResult('A', from_, to_);
    r.Append(Pop());
    _stack.Append(r);
}

wRef TclEngine::TclScan(const char* from_, const char* to_)
{
    _stack.Append("T");
    TclTokenizer tokenizer (from_, to_, *this);
    tokenizer.TokenizeScript();
    return _stack;
}

wRef TclEngine::TclExpr(const char* from_, const char* to_)
{
    TclExprScanner scanner (from_, to_, *this);
    return scanner.Evaluate();
}

/////////////////////////////////////////////////////////////////////////////

wRef wTclCmd(wCallFrame& cf_)
{
    static const char* cmds[] = {
        "parse", "expr", "list", "quote", 0
    };

    switch (cf_.Lookup(cmds))
    {

    case 0: // parse str ?num? -> tree
        {
            wRef input = ++cf_;
            int num = 0;
            cf_ >> num;

            const char* start = input;
            TclEngine engine (num != 0 ? start : 0);
            return engine.TclScan(input, start + input.Length());
        }
    case 1: // expr str ?num? -> tree
        {
            wRef input = ++cf_;
            int num = 0;
            cf_ >> num;

            const char* start = input;
            TclEngine engine (num != 0 ? start : 0);
            return engine.TclExpr(input, start + input.Length());
        }
    case 2: // list str -> string
        {
            wRef input = ++cf_;

            const char* start = input;
            return ListItem(input, start + input.Length());
        }
    case 3: // quote str ?unbalanced? -> string
        {
            wRef input = ++cf_;
            int num = 0;
            cf_ >> num;

            const char* start = input;
            return QuoteItem(input, start + input.Length(), num != 0);
        }
    }

    return wRef ();
}
